# !/usr/bin/env python
__author__ = "Danelle Cline"
__copyright__ = "Copyright 2022, MBARI"
__credits__ = ["MBARI"]
__license__ = "GPL"
__maintainer__ = "Duane Edgington"
__email__ = "duane at mbari.org"
__doc__ = '''

Main entry point for deepsea-ai

@author: __author__
@status: __status__
@license: __license__
'''

import click
from datetime import datetime
from pathlib import Path
import sys
import numpy as np
from urllib.parse import urlparse

from deepsea_ai.commands import upload_tag, config, process, train, bucket
from deepsea_ai.database import api, queries
from deepsea_ai import __version__

# example s3 buckets for help
example_input_process_s3 = 's3://902005-video-in-dev'
example_output_process_s3 = 's3://902005-tracks-out-dev'
example_input_train_s3 = 's3://902005-training-dev'
example_output_train_s3 = 's3://902005-model-checkpoints-dev'


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(
    __version__,
    '-V', '--version',
    message=f'%(prog)s, version %(version)s'
)
def cli():
    """
    Process deep sea video in AWS from a command line.
    """
    pass


@cli.command(name="ecsprocess")
@click.option('--endpoint', required=False, help='Set to the database endpoint to check if any video has already '
                                                 'been processed and loaded before sending off a job')
@click.option('-u', '--upload', is_flag=True, default=False,
              help='Set option to upload local video files to S3 bucket')
@click.option('--clean', is_flag=True, default=True,
              help='Clean up unused artifact (e.g. video) from s3 after processing')
@click.option('-i', '--input', type=str, default="/Volumes/M3/mezzanine/DocRicketts/2022/02/1423/",
              help='Path to the folder with video files to upload. These can be either mp4 or mov files that '
                   'ffmpeg '
                   'understands.')
@click.option('--cluster', type=str, default='lonny33k',
              help='Name of the cluster to use to batch process.  This must correspond to an available Elastic '
                   'Container Service cluster.')
@click.option('--job', type=str, default='lonny33k',
              help='Name of the job, e.g. DiveV4361 benthic outline')
def batchprocess_command(endpoint, upload, clean, cluster, job, input):
    """
     (optional) upload, then batch process in an ECS cluster
    """
    database = None
    if endpoint:
        database = api.DeepSeaAIClient(endpoint)

    input_path = Path(input)
    resources = config.get_resources(cluster)
    user_name = config.get_username()

    for v in videos:
        loaded = False
        if database:
            # Check if the video has already been loaded by looking it up by the media name per this job name
            medias = database.execute(queries.GET_MEDIA_IN_JOB,
                                      processing_job_name=f"{resources['processor']}-“{job_name}”",
                                      media_name=v.name)

            # Found a media in the job as keyed by the processing name, so assume that this was already processed
            if len(medias['data']['mediaInJob']) > 0:
                loaded = True

        if upload and not loaded:
            upload_tag.run([v], resources['VIDEO_BUCKET'], user_name)

        if not loaded:
            process.batch_run(resources, v, job, user_name, clean)
        else:
            print(f'Video {v.name} has already been processed and loaded...skipping')


@cli.command(name="process")
@click.option('-u', '--upload', is_flag=True, default=True,
              help='Set option to upload local video files to S3 bucket')
 # this might be cleaner as an enum but click does not support that fully yet
@click.option('--tracker', default='deepsort',
              help='Tracking type: deepsort or strongsort')
@click.option('-i', '--input', type=str, required=True,
              help='Path to the folder with video files to upload. These can be either mp4 or mov files that '
                   'ffmpeg understands.')
@click.option('--input-s3', type=str, required=True,
              help=f'Path to the s3 bucket with video files. These can be either mp4 or '
                                               f'mov files that ffmpeg understands, e.g. s3://{example_input_process_s3}')
@click.option('--output-s3', type=str, required=True,
              help=f'Path to the s3 bucket to store the output, e.g. s3://{example_output_process_s3}')
@click.option('-m', '--model-s3', type=str, default=process.default_model_s3,
              help='S3 location of the trained model tar gz file - must contain a model.tar.gz file with a valid YOLOv5 '
                   'Pytorch model.')
@click.option('-c', '--config-s3', type=str, help='S3 location of tracking algorithm config yaml file')
@click.option('--model-size', type=click.INT, default=640, help='Size of the model, e.g. 640 or 1280')
@click.option('--conf-thres', type=click.FLOAT, default=.01, help='Confidence threshold for the model')
@click.option('-s', '--save-vid', is_flag=True, default=False,
              help='Set option to output original video with detection boxes overlaid.')
@click.option('--instance-type', type=str, default='ml.p3.xlarge', help='AWS instance type, e.g. ml.p2.xlarge, ml.p3.2xlarge, ml.p3.8xlarge, ml.p3.16xlarge, ml.p4d.24xlarge')
def process_command(upload, tracker, input, input_s3, output_s3, model_s3, config_s3, model_size, conf_thres,
                    save_vid, instance_type):
    """
     (optional) upload, then process video with a model
    """
    if instance_type == 'ml.p2.xlarge':
        raise Exception(f'{instance_type} too small. Choose ml.p3.2xlarge or better')

    input_path = Path(input)
    input_s3 = urlparse(input_s3)
    output_s3 = urlparse(output_s3)
    model_s3 = urlparse(model_s3)

    # get tags to apply to the resources for cost monitoring
    tags = config.get_tags()

    # create the buckets
    print(f'Creating buckets')
    bucket.create(input_s3, tags)
    bucket.create(output_s3, tags)

    if upload:
        videos = config.check_videos(input_path)
        upload_tag.video_data(videos, input_s3, tags)
 
    # insert the datetime prefix to make a unique key for the output
    now = datetime.utcnow()
    prefix = now.strftime("%Y%m%dT%H%M%SZ")
    output_unique_s3 = urlparse(f"s3://{output_s3.netloc}{output_s3.path}/{prefix}/")

    volume_size_gb = 10
    
    process.script_processor_run(input_s3, output_unique_s3, model_s3, model_size, volume_size_gb, instance_type,
                                 config_s3, save_vid, conf_thres, tracker)


@cli.command(name="upload")
@click.option('-i', '--input', type=click.Path(exists=True, file_okay=False, dir_okay=True), required=True,
              help='Path to the folder with video files to upload. These can be either mp4 or mov files that '
                   'ffmpeg understands.')
@click.option('--s3', type=str, help='S3 bucket to upload to, e.g. s3://902005-video-in-dev', required=True)
def upload_command(input, s3):
    """
    Upload videos
    """
    input_path = Path(input)
    input_s3 = urlparse(s3)
    tags = config.get_tags()
    bucket.create(input_s3, tags)
    videos = config.check_videos(Path(input))

    upload_tag.video_data(videos, input_s3, tags)


@cli.command(name="train")
@click.option('--images', type=str, required=True,
              help='Path to compressed training images. Images should be put into a tar.gz file '
                   'that decompress into images/train images/val and optionally images/test. This compressed file should not include any archive artifacts,'
                   ' e.g. ._ files. For example, compress to avoid archives in Mac OSX with'
                   ' COPYFILE_DISABLE=1 tar -czf images.tar.gz images/.  ')
@click.option('--labels', type=str, required=True,
              help='Path to the compressed training labels in YOLO format. Labels should be put into a tar.gz file '
                   'that decompress into labels/. This compressed file should not include any archive artifacts,'
                   ' e.g. ._ files. For example, compress to avoid archives in Mac OSX with'
                   ' COPYFILE_DISABLE=1 tar -czf labels.tar.gz labels/.  ')
@click.option('--label-map', type=str, required=True,
              help='Path to a simple text file with the yolo names in the sorted order of the training label indexes')
@click.option('--input-s3', type=str, required=True,
              help=f'Path to the s3 bucket to save the training data, e.g. {example_input_train_s3} or {example_input_train_s3}')
@click.option('--output-s3', type=str, required=True,
              help=f'Path to the s3 bucket to store the output, e.g. {example_output_train_s3} or {example_output_train_s3}')
@click.option('--resume', type=bool, default=False, help="Resume training from previous run")
@click.option('--model', type=str, default='yolov5x', help=f"Model choice: {','.join(train.models)} " )
@click.option('--epochs', type=int, default=2, help='Number of epochs. Default 2.')
@click.option('--batch-size', type=int, default=2, help='Batch size. Default 2.')
@click.option('--instance-type', type=str, default='ml.p3.2xlarge', help='AWS instance type, e.g. ml.p3.2xlarge, ml.p3.8xlarge, ml.p3.16xlarge, ml.p4d.24xlarge')
def train_command(images, labels, label_map, input_s3, output_s3, resume, model, epochs, batch_size, instance_type):
    """
     (optional) upload training data, then train a YOLOv5 model
    """
    if instance_type == 'ml.p2.xlarge':
        raise Exception(f'{instance_type} too small for model {model}. Choose ml.p3.2xlarge or better')

    image_path = Path(images)
    label_path = Path(labels)
    name_path = Path(label_map)

    # strip off any s3 prefixes
    input_s3 = urlparse(input_s3)
    output_s3 = urlparse(output_s3)

    data = [image_path, label_path, name_path]

    # get tags to apply to the resources for cost monitoring
    tags = config.get_tags()

    # create the buckets
    bucket.create(input_s3, tags)
    bucket.create(output_s3, tags)

    # upload and return the final bucket prefix to the training data and its total size
    input_training, size_gb = upload_tag.training_data(data, input_s3, tags)

    # guess on how much volume is needed per each GB plus the size for the checkpoints
    volume_size_gb = int(2*size_gb + 50)

    # insert the datetime prefix to make a unique key for the outputs
    now = datetime.utcnow()
    prefix = now.strftime("%Y%m%dT%H%M%SZ")

    if resume: # resuming from previous bucket, so no need to set prefix
        ckpts_s3 = urlparse(f"s3://{output_s3.netloc}{output_s3.path.lstrip('/')}")
    else:
        ckpts_s3 = urlparse(f"s3://{output_s3.netloc}{output_s3.path.rstrip('/')}/{prefix}/checkpoints/")
    model_s3 = urlparse(f"s3://{output_s3.netloc}{output_s3.path.rstrip('/')}/{prefix}/models/")

    # train
    train.yolov5(data, input_training, ckpts_s3, model_s3, epochs, batch_size, volume_size_gb,  model, instance_type)


@cli.command(name="package")
@click.option('--s3', type=str,  required=True,
              help=f"Bucket with the model checkpoints, e.g. s3://{example_output_train_s3}/20220821T005204Z"
              )
def package_command(s3):
    train.package(urlparse(s3))

@cli.command(name="split")
@click.option('-i', '--input', type=str, required=True,
              help='Path to the root folder with images and labels, organized into labels/ and images/ folders files to split')
@click.option('-o', '--output', type=str, required=True,
              help='Path to the root folder to save the split, compressed files. If it does not exist, it will be created.')
def split_command(input:str, output:str):
    """
    split data into train/val/test sets randomly per the following percentages 85%/10%/5%
    """
    input_path = Path(input)
    output_path = Path(output)

    if not output_path.exists():
        output_path.mkdir(parents=True)

    paths = [input_path / 'labels', input_path / 'images']

    exists = [not p.exists() for p in paths]
    if any(exists):
        print (f'Error: one or more {paths} missing')
        return

    train.split(Path(input), Path(output))

if __name__ == '__main__':
    cli()



# !/usr/bin/env python
__author__ = "Danelle Cline, Duane Edgington"
__copyright__ = "Copyright 2022, MBARI"
__credits__ = ["MBARI"]
__license__ = "GPL"
__maintainer__ = "Duane Edgington"
__email__ = "duane at mbari.org"
__doc__ = '''

Configuration helper to setup defaults and fetch cloud configuration

@author: __author__
@status: __status__
@license: __license__
'''

import boto3
import datetime as dt
import os
from botocore.exceptions import ClientError
from pathlib import Path
from typing import List

default_training_prefix = 'training'

def get_role():
    """
    Get the user role or default to an MBARI generated one
    :return
    """
    if 'SAGEMAKER_ROLE' not in os.environ:
        raise Exception(f"SAGEMAKER_ROLE must be set in your environment variables")
    return os.environ['SAGEMAKER_ROLE']

def get_account() -> str:
    """
    Get the account number associated with this user
    :return:
    """
    account_number = boto3.client('sts').get_caller_identity()['Account']
    print(f'Found account {account_number}')
    return account_number


def get_username() -> str:
    """
    Get the user name using IAM; if IAM is not configured, this will default to the root user which may be the case
    for a new AWS account
    :return:
    """
    try:
        iam = boto3.client('iam')
        if 'UserName' in iam.get_user()["User"]:
            user_name = iam.get_user()["User"]["UserName"]
        else:
            user_name = 'root'
    except ClientError as e:
        # The user_name may be specified in the Access Denied message...
        user_name = e.response["Error"]["Message"].split(" ")[-1]

    return user_name


def get_tags() -> dict:
    """
    Configure tag dictionary to associate with a job.
    This is useful for cost accounting and general reporting
    :return:
    """
    user_name = get_username()
    deletion_date = (dt.datetime.utcnow() + dt.timedelta(days=90)).strftime('%Y%m%dT%H%M%SZ')
    tag_dict = [{'Key': 'mbari:project-number', 'Value': '902005'},
                {'Key': 'mbari:owner', 'Value': user_name},
                {'Key': 'mbari:description', 'Value': 'test pipeline'},
                {'Key': 'mbari:customer-project', 'Value': '902005'},
                {'Key': 'mbari:stage', 'Value': 'test'},
                {'Key': 'mbari:application', 'Value': 'detection'},
                {'Key': 'mbari:deletion-date', 'Value': deletion_date},
                {'Key': 'mbari:created-by', 'Value': user_name}]
    return tag_dict


def check_videos(input_path: Path) -> List[Path]:
    """
     Check for videos with acceptable suffixes and return the Paths to them
    :param input_path: input path to search (non-recursively)
    :return:
    """
    vid_formats = ['.mov', '.avi', '.mp4', '.mpg', '.mpeg', '.m4v', '.wmv', '.mkv']  # acceptable video suffixes
    files = sorted(input_path.glob('**/*'))
    videos = [x for x in files if x.suffix.lower() in vid_formats and '._' not in x.name]
    num_videos = len(videos)
    assert (num_videos > 0), "No videos to process"
    video_paths = [Path(x) for x in videos]
    return video_paths


def get_resources(stack_name: str) -> dict:
    """
    Get resources relevant to the pipeline from the stack name; see deepsea-ai/cluster/stacks
    :param stack_name:
    :return:
    """
    client_cf = boto3.client('cloudformation')
    client_ecs = boto3.client('ecs')

    stack_resources = client_cf.list_stack_resources(StackName=stack_name)
    resources = {'CLUSTER': stack_name}

    # fetch the PROCESSOR environment variable for the single task in the stack; the job is keyed uniquely to it
    key = ['PROCESSOR', 'TRACK_QUEUE', 'VIDEO_QUEUE', 'DEAD_QUEUE', 'TRACK_BUCKET', 'VIDEO_BUCKET']
    for r in stack_resources['StackResourceSummaries']:
        if 'AWS::ECS::TaskDefinition' in r['ResourceType']:
            arn = r['PhysicalResourceId']
            task_def = client_ecs.describe_task_definition(taskDefinition=arn)
            environment = task_def['taskDefinition']['containerDefinitions'][0]['environment']
            for e in environment:
                for k in key:
                    if k in e['name']:
                        resources[k] = e['value']
            break
    print(resources)
    return resources



# !/usr/bin/env python
__author__ = "Danelle Cline"
__copyright__ = "Copyright 2022, MBARI"
__credits__ = ["MBARI"]
__license__ = "GPL"
__maintainer__ = "Duane Edgington"
__email__ = "duane at mbari.org"
__doc__ = '''

Bucket utilities

@author: __author__
@status: __status__
@license: __license__
'''
import logging
import boto3
import numpy as np
from botocore.exceptions import ClientError


def create(bucket:tuple, tags: dict):
    """Create an S3 bucket
    :param bucket: Bucket to create
    :param tags: Tags to assign to the bucket
    :return: True if bucket created, else False
    """

    # Create bucket
    try:
        session = boto3.session.Session()
        s3_client = boto3.client('s3', region_name=session.region_name)
        location = {'LocationConstraint': session.region_name}
        s3_client.create_bucket(Bucket=bucket.netloc,
                                CreateBucketConfiguration=location)

        try:
            s3_client.put_bucket_tagging(Bucket=bucket.netloc, Tagging={'TagSet': tags})
        except Exception as error:
            raise error
    except ClientError as e:
        if e.response['Error']['Code'] != "BucketAlreadyOwnedByYou":
            logging.error(e)
        return False
    return True


def download(bucket_name:str, prefix: str):
    """Download an object by its prefix from a bucket
    :param bucket_name: Bucket to create
    :param prefix: Prefix to download
    :return: True if bucket created, else False
    """

    # Create bucket
    try:
        session = boto3.session.Session()
        s3_client = boto3.client('s3', region_name=session.region_name)
        location = {'LocationConstraint': session.region_name}
        s3_client.create_bucket(Bucket=bucket_name,
                                CreateBucketConfiguration=location)

        try:
            s3_client.put_bucket_tagging(Bucket=bucket_name, Tagging={'TagSet': tags})
        except Exception as error:
            raise error
    except ClientError as e:
        if e.response['Error']['Code'] != "BucketAlreadyOwnedByYou":
            logging.error(e)
        return False
    return True

def size(bucket:tuple) -> int:
    """
    Get the total size in GB of the bucket.
    :param bucket: Bucket name to searc
    :return: Total size in gigabytes
    """
    size_gb = 0
    session = boto3.session.Session()
    s3 = session.resource('s3')
    b = s3.Bucket(bucket.netloc)

    for object in b.objects.filter(Prefix=bucket.path.split('/')[-1]):
        folder = object.key.split('/')[0]
        size_gb += object.size

    return np.round(size_gb/1e9)


GET_TRACKS_FOR_MEDIA = """
       query getTracksForMedia($media_id: Int!) {
           tracksByMediaId(media_id: $media_id) {
               track_uuid
               max_concept
               start_frame_number
               end_frame_number
           }
       }
       """

GET_MEDIA_IN_JOB = """
query getMediaInJob($processing_job_name: String!, $media_name: String!) {
    mediaInJob(processing_job_name: $processing_job_name, media_name: $media_name) 
    {
        name
        uuid
    }
}
"""


# !/usr/bin/env python
__author__ = "Danelle Cline"
__copyright__ = "Copyright 2022, MBARI"
__credits__ = ["MBARI"]
__license__ = "GPL"
__maintainer__ = "Danelle Cline"
__email__ = "dcline at mbari.org"
__doc__ = '''

Runs deepsort tracking algorithm on YOLOv5 models

The speed this runs depends on many factors including: 
1) the size of the detection model. 
2) the GPU and CPU it is run on

@author: __author__
@status: __status__
@license: __license__
'''

import click
import datetime
import boto3
import tempfile
import json
import signal
import yaml
import tarfile
import os
import shutil
import subprocess
import sys
from pathlib import Path
from urllib.parse import urlparse
import uuid
from pipeline import queue_processor
from pipeline import __version__

# If running in AWS, we must define the inputs/outputs per the spec
from pipeline.data_models import generate_uuids, parse_events

if 'AWS_CONTAINER_CREDENTIALS_RELATIVE_URI' in os.environ:
    default_input = '/opt/ml/processing/input'
    default_output = '/opt/ml/processing/output'
    job_config_path = Path('/opt/ml/config/')
    sys.path.insert(0, '/app/Yolov5_DeepSort_Pytorch')
else:
    default_input = Path(__file__).parent.parent / 'test' / 'in'
    default_output = Path(__file__).parent.parent / 'test' / 'out'
    sys.path.insert(0, str(Path(__file__).parent.parent))
    job_config_path = None

print(default_output)
print(default_input)
stop_flag = False

# A known pretrained model
default_model_s3 = 's3://902005-public/models/yolov5x_mbay_benthic_model.tar.gz'  # yolov5 model

MAX_TIMEOUT_SECS = 3*60*60 # maximum time a file should take to process is 3 hours

@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(
    __version__,
    '-V', '--version',
    message=f'%(prog)s, version %(version)s'
)
def cli():
    """
    DeepSort Detect and Track
    """

@cli.command(name="dettrack")
@click.option('-c', '--config-s3', type=click.STRING, help='Location of deepsort tracking algorithm config yaml file')
@click.option('-i', '--input', type=click.STRING, default=default_input,
              help='Path to the input path with video_path files. These can be either mp4 or mov files that ffmpeg '
                   'understands.')
@click.option('-o', '--output', type=click.STRING, default=default_output,
              help='Path to the output to save the results')
@click.option('--conf-thres', type=float, default=0.01, help='object confidence threshold')
@click.option('--iou-thres', type=float, default=0.5, help='IOU threshold for NMS')
@click.option('--save-vid', is_flag=True, help='save video_path tracking results')
@click.option('--max-det', type=int, default=1000, help='maximum detections per image')
@click.option('-m', '--model-s3', type=str, default=default_model_s3,
              help='S3 path to the trained model tar gz file - must contain a valid YOLOv5 Pytorch model. ')
@click.option('--model-size', type=click.INT, default=640, help='Size of the model, e.g. 640 or 1280')
@click.option('--debug', is_flag=True, help='Debugging flag. Skips processing and downloading of the model.')
def process_command(config_s3, conf_thres, iou_thres, input, output, model_size, model_s3, save_vid, max_det, debug):
    """
    Process a collection of videos either from an input folder, or from a sqs queue
    """
    signal.signal(signal.SIGTERM, sigterm_handler)

    print(f'{__version__}')
    print(f'Processing videos in {input}. ')
    input_path = Path(input)
    output_path = Path(output)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    processor = queue_processor.QueueProcessor(input_path, output_path)

    # download and setup the model once in a temp dir
    download_fini = False
    with tempfile.TemporaryDirectory() as model_temp_dir:

        # deepsort track.py requires separate input/output so let's create that here
        with tempfile.TemporaryDirectory() as in_tmp_dir:
            in_tmp_path = Path(in_tmp_dir) / 'in'; in_tmp_path.mkdir()
            out_tmp_path = Path(in_tmp_dir) / 'out'; out_tmp_path.mkdir()

            while processor.has_video() and not stop_flag:

                try:
                    # start fetching new video
                    video = processor.next(out_tmp_path, debug)

                    if video is None:
                        break

                    # only download the model if we have a video to process
                    if not debug and not download_fini:
                        download_fini = True
                        model_path, model_name, config_path = download_config(model_s3, config_s3, Path(model_temp_dir))

                    # output tar to save the results to
                    out_tar_path = Path(output_path) / f'{video.stem}.tracks.tar.gz'
                    out_tar_path.parent.mkdir(parents=True, exist_ok=True)

                    try:
                        start_utc = datetime.datetime.utcnow()

                        if not debug:
                            cmd_track = f'python3 track.py --yolo_model {model_path} ' \
                                        f'--source {video.input_path} ' \
                                        f'--output {in_tmp_path} ' \
                                        f'--imgsz {model_size} ' \
                                        f'--conf-thres {conf_thres} ' \
                                        f'--iou-thres {iou_thres} ' \
                                        f'--max-det {max_det} ' \
                                        f'--save-txt ' \
                                        f'--agnostic-nms '
                            if config_path:
                                cmd_track += f'--config_deepsort {config_path} '
                            if save_vid:
                                cmd_track += '--save-vid '
                        else:
                            cmd_track = [f"echo track {input_path} && sleep 60"]

                        track_proc = subprocess.Popen(cmd_track, shell=True)
                        track_proc.wait(timeout=MAX_TIMEOUT_SECS)
                        if track_proc.returncode == 0 or debug:
                            video.complete()

                    except Exception as ex:
                        print(f'Exception {ex}')
                    finally:

                        # copy any configuration files to further downstream data loading/processing
                        if job_config_path and job_config_path.exists():
                            for c in job_config_path.glob('*.json'):
                                shutil.copy2(c.as_posix(), in_tmp_dir)
                        else:
                            # create a new job file in the temp_dir with job metadata
                            with open(f"{in_tmp_path.as_posix()}/processingjobconfig.json", "w", encoding="utf-8") as j:
                                image_uri = os.environ['IMAGE_URI'] if 'IMAGE_URI' in os.environ else "mbari/deepsort-yolov5:latest"
                                processor_name = os.environ['PROCESSOR'] if 'PROCESSOR' in os.environ else "deepsort-yolov5"
                                json.dump({
                                      "UserName": video.user_name,
                                      "ProcessingJobName": f"{processor_name}-{video.job_name}",
                                      "AppSpecification": {
                                          "ImageUri": image_uri,
                                          "ContainerArguments": sys.argv
                                      }
                                    }, j)
                        # convert the yolo output to a more friendly json output
                        # yolo output is a simple .txt file with the same file prefix as the video_path,
                        # e.g. myvideo.mov output is myvideo.txt
                        yolo_results = in_tmp_path / f'{video.stem}.txt'

                        # handle missing data
                        if yolo_results.exists():
                            events_sans_uuids = parse_events(yolo_results.as_posix())
                            events = generate_uuids(events_sans_uuids)

                            # save events aggregated by frame in the same format as deepsea-track to simplify loading
                            visual_events = []
                            last_frame = -1
                            for e in events:
                                if last_frame == -1:
                                    last_frame = e['frameNum']

                                if e['frameNum'] > last_frame:
                                    with open(f"{in_tmp_dir}/f{last_frame:06}.json", 'w', encoding='utf-8') as f:
                                        json.dump(["visualevents", visual_events], f)
                                        visual_events = []

                                last_frame = e['frameNum']

                                visual_events.append(["visualevent",
                                                          {
                                                              'bounding_box': {
                                                                  'height': e['height'],
                                                                  'width': e['width'],
                                                                  'x': e['x'],
                                                                  'y': e['y']
                                                              },
                                                              'class_name': e['name'],
                                                              'confidence': e['confidence'],
                                                              'frame_num': e['frameNum'],
                                                              'occlusion': e['occlusion'],
                                                              'surprise': e['surprise'],
                                                              'track_uuid': e['track_uuid'],
                                                              'uuid': str(uuid.uuid4()),
                                                          }
                                                          ])

                                if len(visual_events) > 0:
                                    with open(f"{in_tmp_path.as_posix()}/f{last_frame:06}.json", 'w', encoding='utf-8') as f:
                                        json.dump(["visualevents", visual_events], f)

                        processor.save(in_tmp_path, out_tar_path, video)
                        total_time = datetime.datetime.utcnow() - start_utc
                        print(f'Total {video.name} processing time {total_time}. Started at {start_utc}')

                        # clean temp input dir and remove output
                        for i in in_tmp_path.glob('*'):
                            i.unlink()
                        for o in out_tmp_path.glob('*'):
                            o.unlink()
                        processor.clean(video)

                except Exception as ex:
                    print(f'System failure exception {ex}')
                    exit(-1)
    print('Done')


def download_config(model_uri: str, track_uri: str, output_dir: Path):
    """
    Downloads and unpacks the model data and optionally a tracker config yaml

    :param model_uri: full path to model bucket object
    :param track_uri: full path to track bucket object or None if using default
    :param output_dir: full path to directory to save the output
    :return:  a Path object with the path to the model, model name, path to the tracker config
    e.g. /tmp/9sdfggg/best.pt, yolov5x, /tmp/9sdfggg/deepsort_config.yaml
    """
    # download the model and uncompress
    parsed_track_config = None
    if track_uri: parsed_track_config = urlparse(track_uri, allow_fragments=False)
    parsed_url_model = urlparse(model_uri, allow_fragments=False)

    print(f'Downloading model {parsed_url_model.netloc}...')
    s3 = boto3.resource('s3')
    bucket = s3.Bucket(parsed_url_model.netloc)
    found_config = False

    try:
        if parsed_url_model.query:
            prefix = parsed_url_model.path.lstrip('/') + '?' + parsed_url_model.query
            objects = bucket.objects.filter(Prefix=prefix)
            if 'model.tar.gz' not in objects.keys():
                raise Exception(f'Cannot find model.tar.gz in {parsed_url_model}')
            s3.Bucket(parsed_url_model.netloc).download_file(f'{prefix}/model.tar.gz', f'{output_dir}/model.tar.gz')
        else:
            s3.Bucket(parsed_url_model.netloc).download_file(f'{parsed_url_model.path[1:]}', f'{output_dir}/model.tar.gz')

        print(f'Unpacking model {parsed_url_model.netloc}...')
        tarfile.open(f'{output_dir}/model.tar.gz').extractall(output_dir)

        # assuming yolov5 model is version 1, read in the model name from the custom_config.yaml
        if (output_dir / 'custom_config.yaml').exists():
            found_config = True
            config_path = output_dir / 'custom_config.yaml'
            with open(config_path) as f:
                config = yaml.safe_load(f)
                model_name = config['model']

        if not found_config:
            raise Exception('Missing a valid configuration file')
            
        # download the tracker config
        if parsed_track_config:
            s3.Bucket(parsed_track_config.netloc).download_file(f'{parsed_track_config.path[1:]}', f'{output_dir}/deepsort_config.yaml')
            return output_dir / 'best.pt', model_name, output_dir / 'deepsort_config.yaml'
        else:
            return output_dir / 'best.pt', model_name, None

    except Exception as ex:
        print(ex)
        exit(-1)


def sigterm_handler(signal, frame):
    print(f'Got SIGTERM {signal}')
    global stop_flag
    stop_flag = True

if __name__ == '__main__':
    cli()



# !/usr/bin/env python
__author__ = "Danelle Cline, Duane Edgington"
__copyright__ = "Copyright 2022, MBARI"
__credits__ = ["MBARI"]
__license__ = "GPL"
__maintainer__ = "Duane Edgington"
__email__ = "duane at mbari.org"
__doc__ = '''

Bucket upload and tagging utility

@author: __author__
@status: __status__
@license: __license__
'''

import botocore
import boto3
import sys
import numpy as np
from pathlib import Path
from urllib.parse import urlparse

from . import config, bucket

def video_data(videos: [], input_s3:tuple, tags:dict):
    """
     Does an upload and tagging of a collection of videos to S3
    :param videos: Array of video files in the input_path to upload
    :param bucket: Base bucket to upload to, e.g. 902005-video-in-dev
    :param bucket: Tags to assign to the video
    :return:
    """

    print("Reduce your upload speed by running ' aws configure set default.s3.max_bandwidth 62MB/s'")

    s3 = boto3.client('s3')
    s3_resource = boto3.resource('s3')

    # upload and tag the video objects individually
    for v in videos:
        prefix_path = v.parent.as_posix().split("Volumes/")[-1] ## get Directory string
        target_prefix = f'{input_s3.path}{prefix_path}/{v.name}'

        # check if the video exists in s3
        try:
            s3_resource.Object(input_s3.netloc, target_prefix).load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                # The video does not exist so upload
                try:
                    with open(v.as_posix(), "rb") as f:
                        print(f'Uploading {v} to s3://{input_s3.netloc}/{target_prefix}...')
                        s3.upload_fileobj(f, input_s3.netloc, target_prefix)
                except: print("error in uploading to s3")
            else:
                raise
        else:
            # the video does exist.
            print(f'Found s3://{input_s3.netloc}/{target_prefix} ...skipping upload')

        try:
            # tag it
            s3.put_object_tagging(Bucket=input_s3.netloc, Key=f'{target_prefix}', Tagging={'TagSet': tags})
        except Exception as error:
            raise error

def training_data(data: [Path], input:tuple, tags:dict):
    """
     Does an upload and tagging of training data to S3
    :param data: Paths to training data to upload
    :param input: Bucket to upload to
    :param bucket: Tags to assign to the video
    :return: Uploaded bucket path, Size in GB of training data
    """

    # upload and tag the video objects individually
    print("Reduce your upload speed by running 'aws configure set default.s3.max_bandwidth 62MB/s'")

    s3 = boto3.client('s3')
    s3_resource = boto3.resource('s3')

    prefix_path = None

    for d in data:
        # arbitrarily pick the first element to form a prefix; it does not matter but can serve as an intuitive
        # way to reference later.
        if not prefix_path:
            prefix_path = d.parent.as_posix().split("Volumes/")[-1].lstrip('/')

        # check if the video exists in s3
        # all of the data needs to be under the same prefix for training
        target_prefix =  f'{prefix_path}/{config.default_training_prefix}/{d.name}'
        try:
            s3_resource.Object(input.netloc, target_prefix).load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                # The data does not exist so upload
                try:
                    with open(d.as_posix(), "rb") as f:
                        print(f'Uploading {d} to s3://{input.netloc}/{target_prefix}...')
                        s3.upload_fileobj(f, input.netloc, target_prefix)
                except Exception as error:
                    print(f"Error {error} uploading to s3")
            else:
                raise
        else:
            # the data already exist so skip over it
            print(f'Found s3://{input.path}/{target_prefix} ...skipping upload')

        try:
            # tag it
            s3.put_object_tagging(Bucket=input.netloc, Key=target_prefix, Tagging={'TagSet': tags})
        except Exception as error:
            raise error

    output = urlparse(f's3://{input.netloc}/{prefix_path}/{config.default_training_prefix}/')
    size_gb = bucket.size(output)

    return output, size_gb



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deepsea_ai',
 'deepsea_ai.commands',
 'deepsea_ai.database',
 'deepsea_ai.pipeline',
 'deepsea_ai.tests']

package_data = \
{'': ['*']}

install_requires = \
['awscli>=1.25.71,<2.0.0',
 'boto3>=1.24.70,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'requests>=2.26.0,<3.0.0',
 'sagemaker>=2.102.0,<3.0.0',
 'tqdm>=4.41.0,<5.0.0']

entry_points = \
{'console_scripts': ['deepsea-ai = deepsea_ai.__main__:cli']}

setup_kwargs = {
    'name': 'deepsea-ai',
    'version': '1.1.1',
    'description': 'DeepSeaAI is a Python package to simplify processing deep sea video in AWS from a command line.',
    'long_description': '[![MBARI](https://www.mbari.org/wp-content/uploads/2014/11/logo-mbari-3b.png)](http://www.mbari.org)\n[![semantic-release](https://img.shields.io/badge/%20%20%F0%9F%93%A6%F0%9F%9A%80-semantic--release-e10079.svg)](https://github.com/semantic-release/semantic-release)\n![license-GPL](https://img.shields.io/badge/license-GPL-blue)\n[![Python](https://img.shields.io/badge/language-Python-blue.svg)](https://www.python.org/downloads/)\n\nDeepSea-AI is a Python package to simplify processing deep sea video in [AWS](https://aws.amazon.com) from a command line. \n\nIt includes reasonable defaults that have been optimized for deep sea video. The goal is to simplify running these algorithms in AWS.\n\nDeepSea-AI currently supports:\n - *Training [YOLOv5](http://github.com/ultralytics/yolov5) object detection models*\n - *Running [YOLOv5](http://github.com/ultralytics/yolov5) detection and tracking pipelines* on video using either:\n   - [DeepSort](https://github.com/mikel-brostrom/Yolov5_DeepSort_Pytorch)\n   - [StrongSort](https://github.com/mikel-brostrom/Yolov5_StrongSORT_OSNet) \n \n\n**Authors**: Danelle Cline, [dcline@mbari.org](mailto:dcline@mbari.org), Duane Edgington, [duane@mbari.org](mailto:duane@mbari.org)\n\n----\n \n## Requirements\n \n* [An AWS account](https://aws.amazon.com)\n* [Python 3.8 or later](https://python.org/downloads/) \n\n\n**After you have setup your AWS account, confirm your AWS Account by listing your s3 buckets**\n\n```\n$ aws --version\n$ aws s3 ls \n```\n\n## Installing\n\nInstall and update using [pip](https://pip.pypa.io/en/stable/getting-started/):\n\n```shell\n$ pip install -U deepsea-ai\n```\n\nFor more details, see the [official documentation](http://docs.mbari.org/deepsea-ai/install).\n\n----\n\n\n## Commands\n\n* [`deepsea-ai train --help` - Train a YOLOv5 model and save the model to a bucket](https://docs.mbari.org/deepsea-ai/commands/train/)\n* [`deepsea-ai process --help` - Process one or more videos and save the results to  a bucket](https://docs.mbari.org/deepsea-ai/commands/process/)\n* [`deepsea-ai ecsprocess --help` - Process one or more videos using the Elastic Container Service and save the results to  a bucket](https://docs.mbari.org/deepsea-ai/commands/process/)\n* [`deepsea-ai split --help` - Split your training data; required before the train command.](https://docs.mbari.org/deepsea-ai/data/) \n* `deepsea-ai -h` - Print help message and exit.',
    'author': 'Danelle Cline',
    'author_email': 'dcline@mbari.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mbari-org/deepsea-ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
